open Common;;
open Types;;
open Term;;
open Equations;;
open Typing_tree;;
open Proj;;
open Substitutions;;

type system = { mutable eqs : eq list;
                mutable subst : subst2 list;
                mutable ttree : typing_tree };;

let subst_system sbs s =
 let foo subst =
  s.eqs <- List.concat (List.map (subst_eq subst) s.eqs);
  let subst2 = subst_to_subst2 subst in
  s.subst <- subst2 :: s.subst;
  s.ttree <- subst_typing_tree subst2 s.ttree
 in
 List.iter foo (List.rev sbs);;

exception Next;;

let step s numopt =
 let (tau, t, tau', terr) =
  match numopt with
     None -> (try List.find decomposible s.eqs with Not_found -> raise Next)
   | Some num -> try let eq = List.nth s.eqs (num - 1) in
                     if decomposible eq
                     then eq
                     else (print_string "This equation cannot be decomposed. Using default behavior.\n";
                           try List.find decomposible s.eqs with Not_found -> raise Next)
                 with Failure "nth" -> (print_string "Invalid argument. Using default behavior.\n";
                                        try List.find decomposible s.eqs with Not_found -> raise Next)
 in
 s.eqs <- remove1 (tau, t, tau', terr) s.eqs;
 print_string "Decomposing: "; print_eq (tau, t, tau', terr);
 let extr = function TypFun (ts,t') -> (ts,t') | _ -> failwith "step-extr" in
 let (l, sigma) = extr tau' in
 let sbs = if l = []
           then [SSimple (t, sigma, [])]
           else if List.length l = 1
           then [SSimple (List.hd l, tau, terr); SSimple (t, sigma, [])]
           else let foo t (tau, terr) = SSimple (t, tau, terr) in
                let n = List.length l in
                List.rev_map2 foo l (duplicate n (fun i (tau, terr) -> (proj_typ i tau, proj_terr i terr)) (tau, terr))
                @ [SSimple (t, sigma, []); SDupl (n, terr)]
 in
 print_subst_list sbs;
 subst_system sbs s;;

let final s =
 let foo = function
    (tau, t, TypVar t', _) -> (TypFun2 ([typ_to_typ2 tau], TypVar2 t), t')
  | _ -> failwith "final-foo"
 in
 let rec loop = function
    ((sigma, t)::r) -> let sbs = SSimple2 (t, sigma) in
                       print_string "Substituting: "; print_subst2 sbs; print_newline ();
                       s.subst <- sbs :: s.subst;
                       s.ttree <- subst_typing_tree sbs s.ttree;
                       loop (List.map (subst_eq2 sbs) r)
  | [] -> ()
 in
 loop (List.map foo s.eqs);;

let proof s =
 let prnb nb = print_string "("; print_int nb; print_string ")" in
 let prnbs = comma_list "" prnb " & " in
 let prinfer nb nbs typing =
  prnb nb; print_string ": ";
  if nbs <> [] then (prnbs nbs; print_string " => ");
  print_typing typing
 in
 let get_nb =
  let compt = ref 0 in
  fun () -> incr compt; !compt
 in
 let rec loop = function
    TTId typing -> begin match typing with
                      ([(x, tau)], TermId y, sigma) when x = y && tau = sigma -> ()
                    | _ -> print_string "Error in typing tree\n";
                           print_string "Axiom:\n"; print_typing typing;
                           failwith "proof-false-id"
                   end;
                   let nb = get_nb () in
                   prinfer nb [] typing;
                   (nb, typing)
  | TTAppl (tm, tns, typing) ->
       let (nbm, typingm) = loop tm in
       let (nbns, typingns) = List.split (List.map loop tns) in
       let proof_ok = match (typingm, typingns, typing) with
          ((envm, m, TypFun2 ([], sigmam)), [(envn, n, taun)], (env, TermAppl (term1, term2), sigma))
           when m = term1 && n = term2 && sigmam = sigma && eq_env env (envm @ envn) -> true
        | ((envm, m, TypFun2 (l, sigmam)), typings, (env, TermAppl (term1, term2), sigma))
           when m = term1 && List.length l = List.length typings && sigmam = sigma ->
            let (envns, ns, tauns) = list_split3 typings in
            eq_env env (envm @ List.concat envns) && l = tauns && List.for_all (fun n -> n = term2) ns
        | _ -> false
       in
       if not proof_ok
       then begin
        print_string "Error in typing\n";
        print_string "Premises:\n"; print_typing typingm; List.iter print_typing typingns;
        print_string "Conclusion:\n"; print_typing typing;
        failwith "proof-false-appl"
       end;
       let nb = get_nb () in
       prinfer nb (nbm::nbns) typing;
       (nb, typing)
  | TTLambda (tm, typing) ->
       let (nbm, typingm) = loop tm in
       begin match (typingm, typing) with
          ((envm, m, taum), (env, TermLambda (x, term), TypFun2 (l, sigma)))
            when m = term && eq_env (env_restr x envm) env && taum = sigma && eq_set_typ2 (env_value x envm) l -> ()
        | _ -> print_string "Error in typing tree\n";
               print_string "Premise:\n"; print_typing typingm;
               print_string "Conclusion:\n"; print_typing typing;
               failwith "proof-false-fun"
       end;
       let nb = get_nb () in
       prinfer nb [nbm] typing;
       (nb, typing)
  | TTForget (tm, tn, typing) ->
       let (nbm, typingm) = loop tm
       and (nbn, typingn) = loop tn in
       begin match (typingm, typingn, typing) with
          ((envm, m, taum), (envn, n, taun), (env, term, tau))
            when taum = tau && term = TermForget (m ,n) && eq_env env (envm @ envn) -> ()
        | _ -> print_string "Error in typing tree\n";
               print_string "Premises:\n"; print_typing typingm; print_typing typingn;
               print_string "Conclusion:\n"; print_typing typing;
               failwith "proof-false-forget"
       end;
       let nb = get_nb () in
       prinfer nb [nbm; nbn] typing;
       (nb, typing)
 in
 print_string "Proof:\n";
 let (_, (_, _, tau)) = loop s.ttree in
 Printf.printf "\nProof rank: %d\n" (rank_typing_tree s.ttree);
 Printf.printf "Final type rank: %d\n" (rank_typ2 tau);;
