open Common;;

type typ_var = string;;

let print_typ_var = print_string;;

type typ =
   TypVar of typ_var
 | TypFun of typ_var list * typ;;

let fresh_typ_var =
 let compt = ref (-1) in
 fun () -> incr compt; "t_" ^ string_of_int !compt;;

let rec fv_typ = function
   TypVar t -> [t]
 | TypFun (ts, tau) -> union ts (fv_typ tau);;

let rec print_typ = function
   TypVar t -> print_typ_var t
 | TypFun (ts, tau) -> if List.length ts > 1 then print_string "(";
                       comma_list "()" print_typ_var "," ts;
                       if List.length ts > 1 then print_string ")";
                       print_string " -> "; print_typ tau;;

type typ2 =
   TypVar2 of typ_var
 | TypFun2 of typ2 list * typ2;;

let rec fv_typ2 = function
   TypVar2 t -> [t]
 | TypFun2 (taus, tau') -> union (unions (List.map fv_typ2 taus)) (fv_typ2 tau');;

let rec typ_to_typ2 = function
   TypVar t -> TypVar2 t
 | TypFun (ts, t) -> TypFun2 (List.map (fun x -> TypVar2 x) ts, typ_to_typ2 t);;

let print_typ2 =
 let rec loop par = function
    TypVar2 t -> print_typ_var t
  | TypFun2 (taus, tau') -> if par then print_string "(";
                            loops true taus;
                            print_string " -> ";
                            loop false tau';
                            if par then print_string ")"
 and loops par l =
  if List.length l > 1 then print_string "(";
  comma_list "()" (loop true) "," l;
  if List.length l > 1 then print_string ")"
 in loop false;;

let print_set_typ2 = List.iter (fun tau -> print_typ2 tau; print_newline ());;

let eq_set_typ2 l1 l2 =
 List.sort compare l1 = List.sort compare l2;;

let inc = function
   0 -> 0
 | n -> n + 1;;

let rec rank_typ2 = function
   TypVar2 _ -> 0
(* | TypFun2 ([], tau') -> rank_typ2 tau' *)
 | TypFun2 ([tau], tau') -> max (inc (rank_typ2 tau)) (rank_typ2 tau')
 | TypFun2 (taus, tau') -> max (inc (max 1 (List.fold_left max 0 (List.map rank_typ2 taus)))) (rank_typ2 tau');;
