open Common;;
open Types;;
open Term;;

type binding = ident_var * typ2;;

let print_binding (x, tau) = print_ident_var x; print_string ":"; print_typ2 tau;;

let fv_binding (_, tau) = fv_typ2 tau;;

let rank_binding (_, tau) = rank_typ2 tau;;

let map_binding f (x, tau) = (x, f tau);;

type env = binding list;;

let print_env env = comma_list "" print_binding ", " env;;

let eq_env l1 l2 =
 List.sort compare l1 = List.sort compare l2;;

let env_restr x env = List.filter (fun (y, _) -> y <> x) env;;

let env_value x env = List.map snd (List.filter (fun (y, _) -> y = x) env);;

let rec env_of_typed_term = function
   TypedTermId (x, t) -> [(x, TypVar2 t)]
 | TypedTermAppl (tm, tn, _) -> union (env_of_typed_term tm) (env_of_typed_term tn)
 | TypedTermLambda (x, tm, _) -> List.filter (fun (y, _) -> x <> y) (env_of_typed_term tm)
 | TypedTermForget (tm, tn, _) -> union (env_of_typed_term tm) (env_of_typed_term tn);;

let fv_env env = unions (List.map fv_binding env);;

let rank_env env = List.fold_left max 0 (List.map rank_binding env);;

let map_env f env = List.map (map_binding f) env;;

type typing = env * term * typ2;;

let typing_of_typed_term tm =
 (env_of_typed_term tm, typed_term_to_term tm, typ_to_typ2 (get_type tm));;

let print_typing (env, m, tau) =
 if env <> [] then (print_env env; print_string " ");
 print_string "|- "; print_term m; print_string " : "; print_typ2 tau; print_newline ()

let fv_typing (env, _, tau) = union (fv_env env) (fv_typ2 tau);;

let rank_typing (env, _, tau) = max (rank_env env) (rank_typ2 tau);;

let map_typing f (env, m, tau) = (map_env f env, m, f tau);;

type typing_tree =
   TTId of typing
 | TTAppl of typing_tree * typing_tree list * typing
 | TTLambda of typing_tree * typing
 | TTForget of typing_tree * typing_tree * typing;;

let rec typed_term_to_typing_tree tm =
 let typing = typing_of_typed_term tm in
 match tm with
    TypedTermId (x, t) -> TTId typing
  | TypedTermAppl (tm, tn, t) -> TTAppl (typed_term_to_typing_tree tm, [typed_term_to_typing_tree tn], typing)
  | TypedTermLambda (x, tm, tau) -> TTLambda (typed_term_to_typing_tree tm, typing)
  | TypedTermForget (tm, tn, tau) -> TTForget (typed_term_to_typing_tree tm, typed_term_to_typing_tree tn, typing);;

let rec fv_typing_tree = function
   TTId typing -> fv_typing typing
 | TTAppl (m, ns, typing) -> union (union (fv_typing_tree m) (unions (List.map fv_typing_tree ns))) (fv_typing typing)
 | TTLambda (m, typing) -> union (fv_typing_tree m) (fv_typing typing)
 | TTForget (m, n, typing) -> union (union (fv_typing_tree m) (fv_typing_tree n)) (fv_typing typing);;

let rec rank_typing_tree = function
   TTId typing -> rank_typing typing
 | TTAppl (m, ns, typing) -> max (max (List.fold_left max 0 (List.map rank_typing_tree ns)) (rank_typing_tree m))
                                 (rank_typing typing)
 | TTLambda (m, typing) -> max (rank_typing_tree m) (rank_typing typing)
 | TTForget (m, n, typing) -> max (max (rank_typing_tree m) (rank_typing_tree n)) (rank_typing typing);;

let map_typing_tree f =
 let rec loop = function
    TTId typing -> TTId (map_typing f typing)
  | TTAppl (m, ns, typing) -> TTAppl (loop m, List.map loop ns, map_typing f typing)
  | TTLambda (m, typing) -> TTLambda (loop m, map_typing f typing)
  | TTForget (m, n, typing) -> TTForget (loop m, loop n, map_typing f typing)
 in loop;;
